import pygame, sys, math, sprite
from pygame.locals import *
from consts import *

try:
	import Numeric as N
	import pygame.surfarray as surfarray
except ImportError:
	raise ImportError, "Numeric and Surfarray are required."

class GelArena:
	def __init__(self, img):
		self.bg = img
		self.nodes = N.zeros((4, 4, 11), N.Float)
		self.nodes[0][0,:] = 450.0
		self.nodes[1][0,:] = 30.0
		self.nodes[3][0,:] = 30.0
		self.nodes[2][0,:] = 610.0

		#try another approach, using an offset table. This is updated in update
		self.voffset = N.zeros((640))

	def bound_sprite(self, sprite):
		bbox = sprite.get_bbox()	
		sprite.onfloor = sprite.onwall = 0
# Floor
		left = bbox[3][0] + sprite.rect.x
		top = bbox[3][1] + sprite.rect.y
		width = bbox[3][2]
		height = bbox[3][3]
		max = -1
		for x in range(left, left+width):
			tmp = self.interpolate(x, 0)
			if tmp <= top + height:
				tmp2 = top + height - tmp
				if max < tmp2: max = tmp2
		if max > -1:
			sprite.onfloor = 1
			sprite.moveyby(-max)
			fv= self.interpolate(left+width/2, 0, 1)
			force = (8.0*sprite.vely + fv)/9.0
			self.set_force(left+width/2, force, 0, 1)
			sprite.vely = force #self.interpolate(left+width/2, 0, 1)
#Roof	
		left = bbox[0][0] + sprite.rect.x
		top = bbox[0][1] + sprite.rect.y
		width = bbox[0][2]
		height = bbox[0][3]
		
		max = -1
		for x in range(left, left+width):
			tmp = self.interpolate(x, 1)
			if tmp >= top:
				tmp2 = tmp - top
				if max < tmp2: max = tmp2
		if max > -1:
			#sprite.stats.balance = 0
			sprite.onfloor = -1
			sprite.moveyby(max)
			fv = self.interpolate(left+width/2, 1, 1)
			force = (8.0*sprite.vely + fv)/9.0
			self.set_force(left+width/2, force, 1, 1)
			sprite.vely = force
				#self.apply_force(left+width/2, force, 1, 1)
				#sprite.vely += force
#Left Wall
		left = bbox[1][0] + sprite.rect.x
		top = bbox[1][1] + sprite.rect.y
		width = bbox[1][2]
		height = bbox[1][3]
		
		max = -1

		for y in range(top, top+height):
			tmp = self.interpolate(y, 3)
			if tmp >= left:
				tmp2 = tmp - left
				if max < tmp2: max = tmp2
		if max > -1:
			sprite.onwall = -1
			sprite.movexby(max)
			fv = self.interpolate(top+height/2, 3, 1)
			force = (8.0*sprite.velx+fv)/9.0
			self.set_force(top+height/2, force, 3, 1)
			sprite.velx = force
				#self.apply_force(top+height/2, -force*4.0, 3)
				#sprite.velx = self.interpolate(top+height/2, 3, 1)
#Right Wall
		max = -1
		for y in range(top, top+height):
			tmp = self.interpolate(y, 2)
			if tmp <= left+width:
				tmp2 = left+width-tmp
				if max < tmp2: max = tmp2
		if max > -1:
			sprite.onwall = 1
			sprite.movexby(-max)
			fv = self.interpolate(top+height/2, 2, 1)
			force = (8.0*sprite.velx+fv)/9.0
			self.set_force(top+height/2, force, 2, 1)
			sprite.velx = force

# Stand on Walls
		left = bbox[3][0] + sprite.rect.x
		top = bbox[3][1] + sprite.rect.y
		width = bbox[ 3][2]
		height = bbox[3][3]
		offset = -1
		for x in range(top + height/2, top + height):
			tmp1 = self.interpolate(x,2)
			tmp2 = self.interpolate(x,3)
			if tmp1 < left + width/2 or tmp2 > left + width/2:
				offset = x - top + height/2
				break

		if offset > -1:
			sprite.onfloor = 1
			sprite.moveyby(-offset)
			sprite.vely = 0

#attack Walls, floor, and Roof
		left = bbox[2][0] + sprite.rect.x
		top = bbox[2][1] + sprite.rect.y
		width = bbox[2][2]
		height = bbox[2][3]
		floormax = -1
		roofmax = -1
		leftmax = -1
		rightmax = -1
		for x in range(left, left+width):
			tmp = self.interpolate(x, 0)
			rtmp = self.interpolate(x,1)
			if tmp <= top + height:
				tmp2 = top + height - tmp
				if floormax < tmp2:floormax = tmp2
			if rtmp >= top:
				tmp2 = rtmp - top
				if roofmax < tmp2: roofmax = tmp2
		for y in range(top, top+height):
			tmp = self.interpolate(y, 2)
			ltmp = self.interpolate(y,3)
			if tmp <= left + width:
				tmp2 = left + width - tmp
				if rightmax < tmp2: rightmax = tmp2
			if ltmp >= left:
				tmp2 = rtmp - left
				if leftmax < tmp2: leftmax = tmp2
		mult = [-1,1][sprite.direction > 0]
		if floormax > -1 and sprite.move.force[1] > 0:
			force = sprite.move.force[1]*STRENGTH[sprite.move.damage]*5.0
			sprite.moveyby(-floormax)
			self.set_force(left+width/2, force, 0, 1)
			force = sprite.move.force[1]*STRENGTH[sprite.move.power]*5.0
			sprite.vely = -force #self.interpolate(left+width/2, 0, 1)
			force = mult*sprite.move.force[0]*STRENGTH[sprite.move.power]*5.0
			sprite.velx = -force #self.interpolate(left+width/2, 0, 1)
		if roofmax > -1 and sprite.move.force[1] < 0:
			force = sprite.move.force[1]*STRENGTH[sprite.move.power]*5.0
			self.apply_force(left+width/2, force)
			sprite.accely -= force
			force = mult*sprite.move.force[0]*STRENGTH[sprite.move.power]*5.0
			sprite.accelx -= force
		if rightmax > -1 and sprite.move.force[0] > 0:
			force = sprite.move.force[1]*STRENGTH[sprite.move.power]*5.0
			self.apply_force(top+height/2, force)
			sprite.accely -= force
			force = mult*sprite.move.force[0]*STRENGTH[sprite.move.power]*5.0
			sprite.accelx -= force
		if leftmax > -1 and sprite.move.force[0] < 0:
			force = sprite.move.force[1]*STRENGTH[sprite.move.power]*5.0
			self.apply_force(top+height/2, force)
			sprite.accely -= force
			force = mult*sprite.move.force[0]*STRENGTH[sprite.move.power]*5.0
			sprite.accelx -= force

		

	def update(self, t):
		cur = 0.0
		while cur + 0.05 < t:
			self.update_step(0.05)
			cur += 0.05
		if t - cur > 0.000001:
			self.update_step(t - cur)

	def update_step(self, t):
		k = 0.2 # spring const
		k2 = 0.005
		kh = 0.5
		b = 0.05 #dampaning const
		for i in range(1,10):
			for j in range(4):
				
				tmp = self.nodes[j][2, i]
				self.nodes[j][2, i] += self.nodes[j][3,i]*4.0 #Jello is not heavy
				tmp = t*(tmp + self.nodes[j][2,i])/2.0#dv = a*dt
				self.nodes[j][0,i] += t*(tmp + self.nodes[j][1,i])/2.0# dx = v*dt
				self.nodes[j][1,i] = tmp
			
				#reset force
				if j == 0:
					spacing = 64
					self.nodes[j][3,i] = -k*(self.nodes[j][0,i]-450) #vert spring
				elif j == 1:
					spacing = 64
					self.nodes[j][3,i] = -k*(self.nodes[j][0,i]-30) #vert spring
				elif j == 2:
					spacing = 48
					self.nodes[j][3,i] = -k*(self.nodes[j][0,i]-610) #horiz spring
				else:
					spacing = 48
					self.nodes[j][3,i] = -k*(self.nodes[j][0,i]-30) #horiz spring
				self.nodes[j][3,i] += - b*self.nodes[j][1,i] #dampening

				#horizontal springs. Each node cannot move horizontally
				spring = self.nodes[j][0,i] - self.nodes[j][0,i-1]
				dist = math.sqrt(spacing*spacing + spring*spring)
				self.nodes[j][3,i] += -spring/dist*kh*(dist-spacing) # left spring
				spring = self.nodes[j][0,i] - self.nodes[j][0,i+1]
				dist = math.sqrt(spacing*spacing + spring*spring)
				self.nodes[j][3,i] += -spring/dist*kh*(dist-spacing) # right spring
			#roof to floor springs
			spring = self.nodes[0][0,i] - self.nodes[1][0, i]
			dist = math.fabs(spring)
			self.nodes[0][3,i] += -spring/dist*k2*(dist-420)
			self.nodes[1][3,i] += spring/dist*k2*(dist-420)
			spring = self.nodes[2][0,i] - self.nodes[3][0, i]
			dist = math.fabs(spring)
			self.nodes[2][3,i] += -spring/dist*k2*(dist-580)
			self.nodes[3][3,i] += spring/dist*k2*(dist-580)


	def apply_force(self, p, f, loc = 0, stat = 3):
		if loc > 1: 
			size = 480
			spacing = 48
		else: 
			size = 640
			spacing = 64
		#if stat == 3: f *= 10.0
		if p > 0 and p < size:
			i = p/spacing
			mult = (p - float(i)*spacing)/spacing
			self.nodes[loc][stat,i] += f*(1-mult)
			self.nodes[loc][stat,i+1] += f*mult
			
	def set_force(self, p, f, loc = 0, stat = 3):
		if loc > 1: 
			size = 480
			spacing = 48
		else: 
			size = 640
			spacing = 64
		if stat == 3: f *= 10.0
		if p > 0 and p < size:
			i = p/spacing
			mult = (p - float(i)*spacing)/spacing
			self.nodes[loc][stat,i] = f*(1-mult)
			self.nodes[loc][stat,i+1] = f*mult

	# interpolate using an nth Lagrange interpolating polynomial
	def interpolate(self, x, loc, stat=0):
		if loc < 2: 
			idx = x/64
			spacing = 64
		else: 
			idx = x/48
			spacing = 48

		sum = 0.0
		if idx == 0: idx = 1
		elif idx > 8: idx = 8
		for k in range(idx-1, idx+3):
			Lu = 1.0
			Ll = 1
			for i in range(idx-1, idx+3):
				if i != k:
					Lu *= x - i*spacing
					Ll *= k*spacing - i*spacing
			sum += self.nodes[loc][stat,k]*Lu/Ll
		return sum

	def bg_blit(self, screen):
		imgarray = surfarray.array2d(self.bg.convert())
		gelcolor = imgarray[0,0]
		self.warp_columns(imgarray, gelcolor, 0)
		imgarray2 = N.transpose(imgarray)
		self.warp_columns(imgarray2, gelcolor, 2)
		surfarray.blit_array(screen,N.transpose(imgarray2))

	def warp_columns(self, imgarray, gelcolor, idx):
		width = len(imgarray[0])
		height = len(imgarray)
		for i in range(height):
			top = 30 - self.interpolate(i, idx+1)
			floor = self.interpolate(i, idx) - width + 30
			top = int(top)
			floor = int(floor)
			offset = floor + top
			if offset > 0:
				step = width/offset
				skip = width - offset*step
				tmp = N.zeros((offset, step+1))
				tmp2 = N.reshape(imgarray[i][:step*offset], (offset, -1))
				tmp[:,1:] = tmp2
				tmp[:,0] = tmp[:,1]
				tmp = N.ravel(tmp)
				if top < 0:
					tmp2 = N.zeros((-top)) + gelcolor
					imgarray[i][:-top] = tmp2
					imgarray[i][-top:] = tmp[:width+top]
				else:
					if floor < 0:
						tmp2 = N.zeros((-floor)) + gelcolor
						imgarray[i][floor:] = tmp2
						imgarray[i][:width+floor-skip] = tmp[top:top+width+floor]
					else:
						if skip > 0 and floor < skip:
							div = skip - floor
							imgarray[i][-div:] = imgarray[i][-skip:width-skip+div]
						else:
							div = 0

						imgarray[i][:width-div] = tmp[top:width+top-div]
			elif offset < 0:
				step = width/(-offset)
				skip = width + step*offset
				if skip > 0:
					tmp = imgarray[i][:-skip]
				else:
					tmp = imgarray[i]
				tmp2 = N.reshape(tmp, (-offset, -1))
				tmp3 = tmp2[:,1:].copy()
				tmp4 = N.ravel(tmp3)
				if top >= 0:
					extra = N.zeros((-floor)) + gelcolor
					if skip > 0:
						imgarray[i][len(tmp4)-top:len(tmp4)-top+skip] = imgarray[i][-skip:].copy()
					imgarray[i][:len(tmp4)-top] = tmp4[top:]
					imgarray[i][floor:] = extra
				else:
					extra1 = N.zeros((-top)) + gelcolor
					if floor >= 0:
						imgarray[i][-top:width-skip] = tmp4[:len(tmp4)-floor]
						imgarray[i][:-top] = extra1
					else:
						extra2 = N.zeros((-floor)) + gelcolor
						if skip > 0:
							imgarray[i][width+floor-skip:width+floor] = imgarray[i][-skip:].copy()

						imgarray[i][:-top] = extra1
						imgarray[i][-top:width+floor-skip] = tmp4
						imgarray[i][floor:] = extra2
			else: #offset == 0, but there may be a shift
				if top < 0:
					extra = N.zeros((-top)) + gelcolor
					imgarray[i][-top:] = imgarray[i][:top].copy()
					imgarray[i][:-top] = extra
				elif top > 0:
					extra = N.zeros((-floor)) + gelcolor
					imgarray[i][:floor] = imgarray[i][-floor:].copy()
					imgarray[i][floor:] = extra
