import pygame, keys, math
import Numeric as N
from moves import *
from consts import *


class Stats:
	def __init__(self):
		self.health = 100
		self.balance = 100

class Sprite(pygame.sprite.Sprite):
	def __init__(self, img, openfile, moves):
		pygame.sprite.Sprite.__init__(self)
		self.stats = Stats()
		self.moves = moves
		self.move = NO_MOVE
		self.onfloor = 0
		self.onwall = 0
		self.in_air = 0
		self.frames = img
		self.image = img.subsurface(pygame.Rect(0, 0, 64, 64))
		self.rect = pygame.Rect(30, 450-64, 64, 64)
		self.x = 30.0
		self.x_old = 30.0
		self.y = 450.0-64.0
		self.y_old = 450.0-64.0
		self.direction = 1
		self.frame = 0
		self.frame_time = 0
		self.mode = STAND
		self.velx, self.vely  = (0.0, 0.0)
		self.accelx, self.accely = (0.0, 0.0)

		#self.bbox = [[[[0,0,0,0]]*5]*8]*10 #N.zeros((10, 8, 5, 4))
		self.bbox = N.zeros((7, 8, 5, 4))
		
		for k in range(7):
			for j in range(8):
				tmp = openfile.readline().rstrip('\n').split()
				for i in range(5):
					tmp2 = tmp[i].split(':')
					self.bbox[k][j][i][0] = int(tmp2[0])
					self.bbox[k][j][i][1] = int(tmp2[1])
					self.bbox[k][j][i][2] = int(tmp2[2])
					self.bbox[k][j][i][3] = int(tmp2[3])

	def change_mode(self, mode):
		self.mode = mode
		self.frame = 0


	def update(self, key_queue, key_state, t, other_sprite):
		if self.stats.balance <= 0: self.change_mode(DAZED)
		elif self.mode == DAZED: self.change_mode(STAND)

		self.x += t*self.velx + 0.5*self.accelx*t*t
		self.y += t*self.vely + 0.5*self.accely*t*t
		self.velx += t*(self.accelx)
		self.vely += t*self.accely

		self.accely = 100 # gravity
		self.accelx = 0

		#check combos
		if self.mode != DAZED: #self.mode == STAND or self.mode == WALK or self.mode == JUMP:
			m = self.moves.check_combos(key_queue, self.direction)
			if m != None:
				if not m.onfloor or m.onfloor and m.onfloor ==  self.onfloor:
					self.change_mode(m.mode)
					mult = [-1,1][self.direction >= 0]
					self.accelx = mult*m.accel[0]
					self.accely = m.accel[1]
					self.move = m
		
		#check movements
		if self.onfloor > 0 and (self.mode == STAND or self.mode == WALK) :
			if key_state[keys.RIGHT]:
				if self.mode == STAND: self.change_mode(WALK)
				self.accelx += 50
			elif key_state[keys.LEFT]:
				if self.mode == STAND: self.change_mode(WALK)
				self.accelx -= 50
			elif self.mode == WALK:
				self.change_mode(STAND)

		self.rect.x = int(self.x)
		self.rect.y = int(self.y)
		

		# terminal velocity (dampening)
		self.accelx -= (self.velx)*0.5
		self.accely -= (self.vely)*0.5
		
		self.frame_time += t
		
		if self.frame_time >= 0.125:
			if self.stats.balance < 100: self.stats.balance += 1
			self.frame_time -= 0.125
	 		self.frame += 1
			if self.frame >= 8:
				if self.mode == JUMP:
					if self.in_air and self.onfloor:
						self.change_mode(STAND)
						self.in_air = 0
					else:
						if self.onfloor == 0: self.in_air = 1
						self.frame = 7
				else:
					self.frame = 0
					if self.mode != STAND and self.mode != WALK and self.mode != DAZED:
						self.change_mode(self.move.end_mode)
						self.move = NO_MOVE
		
		if self.mode == STAND or self.mode == WALK:
			self.direction = other_sprite - self.x
		self.image = pygame.transform.flip(self.frames.subsurface(pygame.Rect(self.frame*64, self.mode*64, 64, 64)), self.direction < 0, 0)

	
	def movexby(self, x):
		self.rect.x += int(x)
		self.x += x
	
	def moveyby(self, y):
		self.rect.y += int(y)
		self.y += y

	def moveby(self, x, y):
		self.rect.move_ip(x,y)
		self.x += x
		self.y +=y
	
	def moveto(self, x,y):
		self.rect.x = x
		self.rect.y = y
		self.x = float(x)
		self.y = float(y)

	def get_bbox(self):
		if self.direction < 0:
			bbox = []
			for box in self.bbox[self.mode][self.frame]:
				bbox.append([64 - box[0] - box[2], box[1], box[2], box[3]])
		else:
			bbox = self.bbox[self.mode][self.frame]

		return bbox
		

def do_collisions(sprite1, sprite2, bbox1, bbox2):
	check_block(sprite1, sprite2, bbox1, bbox2)
	check_block(sprite2, sprite1, bbox2, bbox1)
	check_attack(sprite1, sprite2, bbox1, bbox2)
	check_attack(sprite2, sprite1, bbox2, bbox1)
	check_bump(sprite1, sprite2, bbox1, bbox2)

def check_block(sprite1, sprite2, bbox1, bbox2):
	check_boxbump([sprite1, sprite2], bbox1[4], bbox2[2])

def check_attack(sprite1, sprite2, bbox1, bbox2):
	if bbox1[2][2] <= 0 or bbox1[2][3] <= 0: return
	
	#check head strike
	attack = check_boxbump([sprite1, sprite2], bbox1[2], bbox2[0])
	if attack != [0,0]:
		print "Head Shot"
		if sprite1.move.damage < DEADLY:
			sprite2.stats.health -= STRENGTH[sprite1.move.damage+1]
		else:
			sprite2.stats.health -= STRENGTH[sprite1.move.damage]
		sprite2.stats.balance -= STRENGTH[sprite1.move.power]
		sprite2.velx += -0.5*sprite2.direction*(sprite1.move.force[0]*STRENGTH[sprite1.move.power]) 
		sprite2.vely += 0.5*(sprite1.move.force[1]*STRENGTH[sprite1.move.power]) 		
		if sprite1.move.force[0]: sprite1.velx = 0
		if sprite1.move.force[1]: sprite1.vely = 0

	#check Feet Strike
	else:
		attack = check_boxbump([sprite1, sprite2], bbox1[2], bbox2[3])
		if attack != [0,0]:
			print "Foot Stomp"
			if sprite1.move.damage > NONE:
				sprite2.stats.health -= STRENGTH[sprite1.move.damage-1]
			else:
				sprite2.stats.health -= STRENGTH[sprite1.move.damage]
			if sprite1.move.power < DEADLY:
				sprite2.stats.balance -= STRENGTH[sprite1.move.power+1]
			else:
				sprite2.stats.balance -= STRENGTH[sprite1.move.power]
				sprite2.velx = -0.5*sprite2.direction*(sprite1.move.force[0]*STRENGTH[sprite1.move.power]) 
				sprite2.vely += 0.5*(sprite1.move.force[1]*STRENGTH[sprite1.move.power]) 
				if sprite1.move.force[0]: sprite1.velx = 0
				if sprite1.move.force[1]: sprite1.vely = 0
	
		#check Body Strike
		else:
			attack = check_boxbump([sprite1, sprite2], bbox1[2], bbox2[1])
			if attack != [0,0]:
				print "Body Shot"
				sprite2.stats.health -= STRENGTH[sprite1.move.damage]
				sprite2.stats.balance -= STRENGTH[sprite1.move.power]
				sprite2.velx = -0.5*sprite2.direction*(sprite1.move.force[0]*STRENGTH[sprite1.move.power]) 
				sprite2.vely += 0.5*(sprite1.move.force[1]*STRENGTH[sprite1.move.power]) 
				if sprite1.move.force[0]: sprite1.velx = 0
				if sprite1.move.force[1]: sprite1.vely = 0

def check_bump(sprite1, sprite2, bbox1, bbox2):
	if sprite1.y < sprite2.y:
		newhead2 = [bbox2[0][0], bbox2[0][1], bbox2[0][2], bbox2[1][1]+bbox2[1][3]/2 - bbox2[0][1]]
		newfoot1 = [bbox1[3][0], bbox1[1][1] + bbox1[1][3]/2, bbox1[3][2], bbox1[3][1]+bbox1[3][3] - bbox1[1][1] - bbox1[1][3]/2]
		standon = check_boxbump([sprite1, sprite2], newfoot1, newhead2)
		if standon[1] < 0:
			sprite1.onfloor = 1
		else:
			sprite1.onfloor = 0
	else:
		newhead1 = [bbox1[0][0], bbox1[0][1], bbox1[0][2], bbox1[1][1]+bbox1[1][3]/2 - bbox1[0][1]]
		newfoot2 = [bbox2[3][0], bbox2[1][1] + bbox2[1][3]/2, bbox2[3][2], bbox2[3][1]+bbox2[3][3] - bbox2[1][1] - bbox2[1][3]/2]
		standon = check_boxbump([sprite2, sprite1], newfoot2, newhead1)
		if standon[1] < 0:
			sprite2.onfloor = 1
		else:
			sprite2.onfloor = 0

	newbody1 = [bbox1[1][0], bbox1[0][1], bbox1[1][2], bbox1[1][1]+bbox1[1][3] - bbox1[0][1]]
	newbody2 = [bbox2[1][0], bbox2[0][1], bbox2[1][2], bbox2[1][1]+bbox2[1][3] - bbox2[0][1]]
	check_boxbump([sprite1, sprite2], newbody1, newbody2)

			
def check_boxbump(sprites, box1, box2):
	rect1 = pygame.Rect(box1[0] + sprites[0].x, box1[1] + sprites[0].y, box1[2], box1[3])
	rect2 = pygame.Rect(box2[0] + sprites[1].x, box2[1] + sprites[1].y, box2[2], box2[3])
	
	rect_diff = rect1.clip(rect2)

	val = [0,0]

	if rect_diff:
		if rect_diff.h < rect_diff.w:
			#do y
			if sprites[0].y < sprites[1].y:
				val[1] = -1
				if sprites[1].onfloor > 0:
					sprites[0].moveyby(-rect_diff.h)
				elif sprites[0].onfloor <  0:
					sprites[1].moveyby(rect_diff.h)
				else:
					sprites[0].moveyby(-rect_diff.h/2)
					sprites[1].moveyby(rect_diff.h/2)
			else:
				val[1] = 1
				if sprites[0].onfloor > 0:
					sprites[1].moveyby(-rect_diff.h)
				elif sprites[1].onfloor <  0:
					sprites[0].moveyby(rect_diff.h)
				else:
					sprites[1].moveyby(-rect_diff.h/2)
					sprites[0].moveyby(rect_diff.h/2)
					
				force = (sprites[0].vely + sprites[1].vely)/2.0
				sprites[0].vely = force/2.0
				sprites[1].vely = force/2.0
		else:
			#do x
			if sprites[0].x < sprites[1].x:
				val[0] = -1
				if sprites[1].onwall > 0:
					sprites[0].movexby(-rect_diff.w)
				elif sprites[0].onwall <  0:
					sprites[1].movexby(rect_diff.w)
				else:
					sprites[0].movexby(-rect_diff.w/2)
					sprites[1].movexby(rect_diff.w/2)
			else:
				val[0] = 1
				if sprites[0].onwall > 0:
					sprites[1].movexby(-rect_diff.w)
				elif sprites[1].onwall <  0:
					sprites[0].movexby(rect_diff.w)
				else:
					sprites[1].movexby(-rect_diff.w/2)
					sprites[0].movexby(rect_diff.w/2)
					
				force = (sprites[0].velx + sprites[1].velx)/2.0
				sprites[0].velx = force/2.0
				sprites[1].velx = force/2.0
	return val

def update_sprites(sprites, key_queue, key_state, t, arena):
	bbox1 = sprites[0].get_bbox()
	bbox2 = sprites[1].get_bbox()
	arena.update(t)
	sprites[0].update(key_queue[0], key_state[0], t, sprites[1].x)
	sprites[1].update(key_queue[1], key_state[1], t, sprites[0].x)
	arena.bound_sprite(sprites[0])
	arena.bound_sprite(sprites[1])
	do_collisions(sprites[0], sprites[1], bbox1, bbox2)

def check_sprite_collision(sprite1, sprite2, bbox1, bbox2, t):
	new_t = t
	if bbox1[2] > bbox2[2]:
		offsetx = bbox2[2]/4.0
	else:
		offsetx = bbox1[2]/4.0
	if bbox1[3] > bbox2[3]:
		offsety = bbox2[3]/4.0
	else:
		offsety = bbox1[3]/4.0
			
	if sprite1.x > sprite2.x:
		x1 = sprite1.x + bbox1[0] + offsetx
		x2 = sprite2.x + bbox2[0] + bbox2[2] - offsetx
	else:
		x1 = sprite1.x + bbox1[0] + bbox1[2] - offsetx
		x2 = sprite2.x + bbox2[0] +  offsetx
	if sprite1.y > sprite2.y:
		y1 = sprite1.y + bbox1[1] + offsety
		y2 = sprite2.y + bbox2[1] + bbox2[3] - offsety
	else:
		y1 = sprite1.y + bbox1[1] + bbox1[3] - offsety
		y2 = sprite2.y + bbox2[1] +  offsety
	bx = sprite1.velx - sprite2.velx
	by = sprite1.vely - sprite2.vely
	ax = sprite1.accelx - sprite2.accelx
	ay = sprite1.accely - sprite2.accely
	
	if ax != 0: dx = bx*bx - 2*ax*(x1-x2)
	else: dx = 0
	
	if ay != 0: dy = by*by - 2*ay*(y1-y2)
	else: dy = 0

	overlap = 0
	if bx != 0 or ax != 0:
		if dx >= 0:
			if ax == 0:
				tx1 = tx2 = (x1-x2)/bx
			else:
				tx1 = (-bx + math.sqrt(dx))/ax
				tx2 = (-bx - math.sqrt(dx))/ax
				curt = -1
				if tx1 > 0 or tx2 > 0:
					if tx1 <= 0: curt = tx2
					elif tx2 <= 0: curt = tx1
					elif tx1 < tx2: curt = tx1
					else: curt = tx2

					newy1 = sprite1.vely*curt + 0.5*sprite1.accely*curt*curt + y1
					newy2 = sprite2.vely*curt + 0.5*sprite2.accely*curt*curt + y2

					if newy1 < newy2:
						overlap = newy2 < newy1 + bbox1[3]
					else:
						overlap = newy1 < newy2 + bbox2[3]
	elif by != 0 or ay != 0:
		if dy >= 0:
			if ay == 0:
				ty1 = ty2 = (y1-y2)/by
			else:
				ty1 = (-by + math.sqrt(dy))/ay
				ty2 = (-by - math.sqrt(dy))/ay
				curt = -1
				if ty1 > 0 or ty2 > 0:
					if ty1 <= 0: curt = ty2
					elif ty2 <= 0: curt = ty1
					elif ty1 < ty2: curt = ty1
					else: curt = ty2

					newx1 = sprite1.velx*curt + 0.5*sprite1.accelx*curt*curt + x1
					newx2 = sprite2.velx*curt + 0.5*sprite2.accelx*curt*curt + x2

					if newx1 < newx2:
						overlap = newx2 < newx1 + bbox1[2]
					else:
						overlap = newx1 < newx2 + bbox2[2]

				if overlap and curt < new_t:
					new_t = curt

	return new_t

def check_wall_collision(pix, val2, vel, accel, wall, edge, direction, t, arena):
	new_t = t
	v1 = arena.interpolate(pix, wall, 1)
	a1 = arena.interpolate(pix, wall, 2)
	f1 = arena.interpolate(pix, wall, 3)
	b = 0.5*v1 - vel
	a = a1 + 2.0*f1 - accel
	
	val1 = arena.interpolate(pix, wall, 0)

	if a != 0: d = b*b - 2*a*(val1-val2)
	else: d = 0
	
	overlap = 0
	if b != 0 or a != 0:
		if d >= 0:
			if a == 0:
				tx1 = tx2 = (val1-val2)/b
			else:
				t1 = (-b + math.sqrt(d))/a
				t2 = (-b - math.sqrt(d))/a
				curt = -1
				if t1 > 0 or t2 > 0:
					if t1 <= 0: curt = t2
					elif t2 <= 0: curt = t1
					elif t1 < t2: curt = t1
					else: curt = t2

					newval1 = 0.5*v1*curt + 0.5*(a1 + 2.0*f1)*curt*curt + val1
					newval2 = vel*curt + 0.5*accel*curt*curt + val2

					if newval1 <= newval2 and curt < new_t:
						new_t = curt
	return new_t
						
